from __future__ import annotations

import abc
from dataclasses import dataclass, field
from pathlib import Path
from typing import Iterable, Self, cast

import marko
import marko.ast_renderer
import marko.inline
import marko.md_renderer

from rooster._config import Config
from rooster._github import PullRequest
from rooster._versions import Version, get_previous_version, parse_versions

VERSION_HEADING_PREFIX = "## "


def new_heading(text: str, level: int = 1):
    return marko.parse("#" * level + " " + text + "\n").children[0]


class Section(abc.ABC):
    @abc.abstractclassmethod
    def new(
        cls, document: Document, element: marko.block.BlockElement, title: str
    ) -> Self:
        raise NotImplementedError

    @classmethod
    def from_elements(
        cls, document: Document, elements: list[marko.block.BlockElement], level: int
    ) -> list[Self]:
        renderer = document.renderer()
        sections = []
        last_section = None
        for element in elements:
            if isinstance(element, marko.block.Heading):
                if element.level < level:
                    last_section = None
                    continue
                if element.level > level:
                    if last_section:
                        last_section.children.append(element)
                    continue

                if element.children:
                    title = renderer.render(element.children[0])
                    entry = cls.new(document=document, element=element, title=title)
                    sections.append(entry)
                    last_section = entry
            elif isinstance(element, marko.block.BlankLine):
                continue
            else:
                if last_section:
                    last_section.children.append(element)

        return sections

    def print(self):
        renderer = self.document.renderer()
        if not hasattr(self.element, "link_ref_defs"):
            self.element.link_ref_defs = {}
        print(renderer.render(self.element), end="")
        for child in self.children:
            if isinstance(child, marko.block.BlankLine):
                print()
            else:
                # TODO: Report this as a bug upstream
                if not hasattr(child, "link_ref_defs"):
                    child.link_ref_defs = {}
                print(renderer.render(child), end="")


@dataclass
class Document:
    document: marko.block.Document

    @classmethod
    def from_markdown(cls, contents: str) -> Self:
        return cls(document=marko.parse(contents))

    @classmethod
    def from_file(cls, file: str | Path) -> Self:
        return cls.from_markdown(Path(file).read_text())

    def to_markdown(self) -> str:
        return self.renderer().render(self.document)

    def to_dict(self) -> dict:
        return marko.ast_renderer.ASTRenderer().render(self.document)

    def renderer(self) -> marko.md_renderer.MarkdownRenderer:
        renderer = marko.md_renderer.MarkdownRenderer()
        # Workaround for issues rendering links in children
        renderer.root_node = self.document
        return renderer

    @classmethod
    def empty(cls) -> Self:
        return cls(document=marko.block.Document())

    def add_child(self, child) -> None:
        self.document.children.append(child)

    def add_children(self, children) -> None:
        self.document.children.extend(children)


@dataclass
class Changelog(Document):
    def versions(self, level: int = 2) -> list[VersionSection]:
        """
        Extract version entries from the changelog.

        Each version is expected to be indicated by a header.

        By default, we expect versions to all be declared as level two headers.
        An alternative level can be requested.
        """
        return VersionSection.from_elements(self, self.document.children, level=level)

    def get_version_section(self, version: Version) -> VersionSection | None:
        for section in self.versions():
            if section.version == version:
                return section
        return None

    def insert_version_section(self, section: VersionSection, level: int = 2) -> None:
        renderer = self.renderer()
        elements = cast(list, self.document.children)
        remove = None
        for i, element in enumerate(tuple(elements)):
            if isinstance(element, marko.block.Heading):
                if element.level == level:
                    version = Version(renderer.render(element.children[0]))

                    # We got to the next heading
                    if remove:
                        remove.append(i)
                        break

                    # Assumes descending versions
                    if version < section.version:
                        break

                    # Replace the existing version
                    if version == section.version:
                        remove = [i]

        if remove:
            for _ in range(remove[0], remove[1]):
                elements.pop(remove[0])
            i = remove[0]

        elements.insert(i, section.element)
        elements.insert(i + 1, marko.block.BlankLine)

        # Insert all the version components
        offset = 0
        for offset, child in enumerate(section.children):
            elements.insert(i + 2 + offset, child)

        if not section.children:
            elements.insert(i + 2, marko.block.BlankLine)
            elements.insert(i + 2, marko.block.HTMLBlock("<!-- No changes -->"))

    @classmethod
    def new(cls) -> Self:
        new = cls.empty()
        new.add_child(new_heading("Changelog"))
        new.add_child(marko.block.BlankLine)
        new.add_child(marko.block.BlankLine)
        return new


@dataclass
class VersionSection(Section):
    document: Document
    element: marko.block.Heading
    title: str
    version: Version
    children: list[marko.parser.element.Element] = field(default_factory=list)

    @classmethod
    def new(
        cls, document: Document, element: marko.block.BlockElement, title: str
    ) -> Self:
        return cls(document, element, title, version=Version(title))

    def sections(self):
        """
        Extract entry sections from the version section.
        """
        # Sections are expected to be nested a single level deeper
        return EntrySection.from_elements(
            self.document, self.children, level=self.element.level + 1
        )

    def all_entries(self):
        entries = []
        for section in self.sections():
            entries.extend(section.entries())
        return entries

    @classmethod
    def from_pull_requests(
        cls,
        document: Document,
        config: Config,
        version: Version,
        pull_requests: Iterable[PullRequest],
        level: int = 2,
    ) -> Self:
        # Initialize the sections dictionary to match the changelog sections config for
        # ordering
        sections = {label: [] for label in config.changelog_sections}

        # De-duplicate pull requests and sort into sections
        for pull_request in set(pull_requests):
            for label in pull_request.labels:
                if label in config.changelog_ignore_labels:
                    break
            else:
                # Iterate in-order of changelog sections to support user-configured precedence
                for label in config.changelog_sections:
                    if label in pull_request.labels:
                        sections[label].append(pull_request)
                        break
                else:
                    sections["__unknown__"].append(pull_request)

        children = []
        for section, pull_requests in sections.items():
            # Omit empty sections
            if not pull_requests:
                continue

            pull_requests = sorted(pull_requests, key=lambda pr: pr.title)

            entry_section = EntrySection.from_pull_requests(
                document=document,
                config=config,
                section=section,
                pull_requests=pull_requests,
            )
            children.append(entry_section.element)
            children.extend(entry_section.children)

        heading_element = new_heading(str(version), level=level)

        return cls(
            document=document,
            element=heading_element,
            title=str(version),
            version=version,
            children=children,
        )

    def as_document(self) -> Document:
        document = Document.empty()
        document.add_child(self.element)
        document.add_child(marko.block.BlankLine)
        document.add_children(self.children)
        return document


@dataclass
class EntrySection(Section):
    document: Document
    element: marko.block.Heading
    title: str
    children: list[marko.parser.element.Element] = field(default_factory=list)

    @classmethod
    def new(
        cls, document: Document, element: marko.block.BlockElement, title: str
    ) -> Self:
        return cls(document, element, title)

    def entries(self):
        """ """
        entries = []
        for element in self.children:
            if isinstance(element, marko.block.List):
                for item in element.children:
                    entries.append(Entry(self.document, item))

        return entries

    @classmethod
    def from_pull_requests(
        cls,
        document: Document,
        config: Config,
        section: str,
        pull_requests: Iterable[PullRequest],
        level: int = 3,
    ) -> Self:
        title = config.changelog_sections.get(section)
        heading = new_heading(title, level)

        lines = []
        for pull_request in pull_requests:
            line = config.change_template.format(pull_request=pull_request)
            lines.append(line)

        return cls(
            document=document,
            element=heading,
            title=title,
            children=(
                [marko.block.BlankLine]
                + marko.parse("\n".join(lines)).children
                + [marko.block.BlankLine]
            ),
        )


@dataclass
class Entry:
    document: Document
    element: marko.block.ListItem

    def content(self) -> str:
        return self.document.renderer().render_children(self.element)


def generate_contributors(pull_requests: list[PullRequest], config: Config) -> str:
    contributors = ""
    authors = {
        pull_request.author
        for pull_request in pull_requests
        if pull_request.author not in config.changelog_ignore_authors
    }
    if authors:
        contributors += "### Contributors\n"
        for author in sorted(authors):
            contributors += f"- [@{author}](https://github.com/{author})\n"
    return contributors


def ensure_spacing(changelog: str) -> str:
    # Sloppily ensure we don't have too much spacing
    while "\n\n\n" in changelog:
        changelog = changelog.replace("\n\n\n", "\n\n")

    # Ensure we always end with a single newline
    return changelog.rstrip("\n") + "\n"


def get_versions_from_changelog(changelog: str) -> list[Version]:
    """
    Get all versions from headings from the changelog
    """

    return parse_versions(
        [
            line[2:].strip()
            for line in changelog.splitlines()
            if line.startswith(VERSION_HEADING_PREFIX)
        ]
    )


def extract_entry(changelog: str, version: Version) -> str | None:
    """
    Extract an entry for the given version from the changelog
    """
    heading = f"{VERSION_HEADING_PREFIX}{version}\n\n"

    versions = get_versions_from_changelog(changelog)
    previous_version = get_previous_version(versions, version)

    # If there are no versions in the file, return `None`
    if not previous_version and heading not in changelog:
        return None

    previous_heading = (
        f"{VERSION_HEADING_PREFIX}{previous_version}\n\n" if previous_version else None
    )

    if heading not in changelog:
        return None

    start = changelog.index(heading)
    end = changelog.index(previous_heading) if previous_heading else len(changelog)
    return changelog[start:end]


def entry_to_standalone(changelog_entry: str, version: Version) -> str:
    """
    Convert an entry from the CHANGELOG file to a standalone entry (omitting the version)
    """
    return changelog_entry.replace(
        f"{VERSION_HEADING_PREFIX} {version}\n",
        f"{VERSION_HEADING_PREFIX} Changes\n<!-- Generated from the CHANGELOG file -->\n",
    )



from __future__ import annotations

from pathlib import Path

import pydantic
import tomllib
from typing_extensions import Self


class Config(pydantic.BaseModel):
    """
    Configuration options and defaults for Rooster.
    """

    major_labels: frozenset[str] = frozenset(["breaking"])
    minor_labels: frozenset[str] = frozenset(["feature"])
    patch_labels: frozenset[str] = frozenset(["fix"])

    changelog_file: str = "CHANGELOG.md"
    changelog_contributors: bool = True
    changelog_sections: dict[str, str] = {
        "breaking": "Breaking changes",
        "feature": "New features",
        "fix": "Bug fixes",
    }
    changelog_ignore_labels: frozenset[str] = frozenset([])
    changelog_ignore_authors: frozenset[str] = frozenset(["dependabot"])

    change_template: (
        str
    ) = "- {pull_request.title} ([#{pull_request.number}]({pull_request.url}))"

    # Paths to files to replace versions at
    version_files: list[Path] = []

    # e.g. "v"
    version_tag_prefix: str = ""

    @pydantic.validator("changelog_sections", always=True)
    def require_unknown_key(cls, value):
        value.setdefault("__unknown__", "Other changes")
        return value

    @classmethod
    def from_directory(cls: type[Self], dirpath: Path) -> Self:
        """
        Load the configuration from a `pyproject.toml` file in a directory.

        If there is no `pyproject.toml` file present, the default config will be returned.
        """
        path = dirpath / "pyproject.toml"
        if not path.exists():
            return cls()
        return cls.from_path(path)

    @classmethod
    def from_path(cls: type[Self], path: Path) -> Self:
        """
        Load the configuration from a `pyproject.toml` file.

        If the file does not exist, an error will be raised.
        """
        pyproject = tomllib.loads(path.read_text())
        section = pyproject.get("tool", {}).get("rooster", {})
        return cls(**section)



"""
Utilities for testing Rooster internals.
"""

import subprocess
from pathlib import Path
from typing import Generator

import pytest


@pytest.fixture
def git_directory(tmp_path: Path) -> Path:
    """
    An empty git project in a temporary directory.

    Configures `git` user for making commits.
    Uses `main` as the default branch name.
    """
    subprocess.check_call(["git", "init", "-b", "main"], cwd=tmp_path)
    subprocess.check_call(
        ["git", "config", "user.email", "test@example.com"], cwd=tmp_path
    )
    subprocess.check_call(["git", "config", "user.name", "Test User"], cwd=tmp_path)
    yield tmp_path


def create_tags(directory: Path, tags: list[str]) -> Generator[str, None, None]:
    """
    Create git tags in the given directory.

    Yields before each tag is created to allow for commits to be made.
    """
    for tag in tags:
        yield tag
        subprocess.check_call(["git", "tag", tag], cwd=directory)


def empty_commit(directory: Path, message: str) -> None:
    """
    Create an empty commit in the given directory.
    """
    subprocess.check_call(
        ["git", "commit", "-m", message, "--allow-empty"], cwd=directory
    )


@pytest.fixture
def mock_pyproject(git_directory: Path) -> Path:
    """
    Creates a mock pyproject.toml file.
    """
    pyproject = git_directory.joinpath("pyproject.toml")
    pyproject.write_text(
        """
        [project]
        name = "test-project"
        version = "0.0.0"
        """
    )
    yield pyproject


@pytest.fixture
def mock_project(git_directory: Path, mock_pyproject: Path) -> Path:
    """
    A basic mock project.
    """
    commit_number = 0
    for i, tag in enumerate(create_tags(git_directory, ["0.1.0", "0.2.0", "0.3.0"])):
        # Generate commits for each tag, increasing the number of commits by one per release
        for inner in range(i + 1):
            commit_number += 1
            empty_commit(
                git_directory, f"Commit {commit_number} should belong to {tag}"
            )

    # Create a bad tag that should be ignored
    create_tags(git_directory, ["foo"])

    empty_commit(
        git_directory, f"Commit {commit_number} should be for the next release"
    )
    commit_number += 1
    empty_commit(
        git_directory, f"Commit {commit_number} should be for the next release"
    )
    yield git_directory


def rooster_command(command: list[str], working_directory: Path | None = None) -> dict:
    process = subprocess.run(
        ["rooster"] + command,
        cwd=working_directory,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    return {
        "exit_code": process.returncode,
        "stdout": process.stdout.decode(),
        "stderr": process.stderr.decode(),
    }



import contextlib
import hashlib
import json
import os

import hishel
import httpx
from httpcore import Request, Response


class GraphQLCacheController(hishel.Controller):
    def __init__(self):
        super().__init__(allow_heuristics=True)
        # Hack support for POST requests as hishel does not allow caching of them
        self._cacheable_methods = ["POST"]

    def is_cachable(self, request: Request, response: Response) -> bool:
        # Allow the cache to be disabled
        if os.environ.get("ROOSTER_NO_CACHE"):
            return False

        # Since GraphQL always returns a 200, we check the response body for errors
        error_in_response_body = False
        try:
            error_in_response_body = json.loads(response.read()).get("errors")
        except Exception:
            pass  # The response cannot be parsed as JSON
        return super().is_cachable(request, response) and not error_in_response_body


@contextlib.contextmanager
def cached_graphql_client():
    """
    A HTTP client with support for caching GraphQL requests.

    The cache will be stored at `$PWD/.cache`.
    """
    with httpx.Client(
        transport=hishel.CacheTransport(
            transport=httpx.HTTPTransport(),
            controller=GraphQLCacheController(),
        ),
        timeout=None,
    ) as client:
        yield client


def _generate_key_patch(request: Request) -> str:
    """
    Patches Hishel's cache key generation to include the request body for GraphQL requests
    which always use the same endpoint unlike traditional REST API requests.
    """
    base_key = _generate_key(request)
    hash = hashlib.new("sha256")
    hash.update(base_key.encode())
    for item in request.stream:
        hash.update(item)
    patched_key = hash.hexdigest()
    return patched_key


# Eek! Hishel doesn't let you change this so I patch it for now instead of implementing
# the whole transport again
# https://github.com/karpetrosyan/hishel/issues/75
_generate_key = hishel._sync._transports.generate_key
hishel._sync._transports.generate_key = _generate_key_patch



import dataclasses
import functools
import os
import re
import shutil
import subprocess
import sys
import textwrap

import httpx

from rooster._cache import cached_graphql_client
from rooster._git import git

TOKEN_REGEX = re.compile(r"Token:\s(.*)")


@dataclasses.dataclass(frozen=True, unsafe_hash=True)
class PullRequest:
    title: str
    number: int
    labels: frozenset[str]
    author: str
    repo_name: str
    repo_owner: str

    @property
    def url(self):
        return (
            f"https://github.com/{self.repo_owner}/{self.repo_name}/pull/{self.number}"
        )


@dataclasses.dataclass(frozen=True, unsafe_hash=True)
class Release:
    id: str
    name: str
    tag: str
    body: str
    draft: bool
    prerelease: bool


@functools.cache
def get_github_token() -> str:
    """
    Retrieve the current GitHub token from the `gh` CLI or `GITHUB_TOKEN` environment variable.

    This function is cached and should only run once invocation of `rooster`.
    """
    if "GITHUB_TOKEN" in os.environ:
        return os.environ["GITHUB_TOKEN"]

    if not shutil.which("gh"):
        print(
            "You must provide a GitHub access token via GITHUB_TOKEN or have the gh CLI"
            " installed."
        )
        raise RuntimeError("Failed to retrieve GitHub token")

    gh_auth_status = subprocess.run(
        ["gh", "auth", "status", "--show-token"], capture_output=True
    )
    output = gh_auth_status.stdout.decode()
    if not gh_auth_status.returncode == 0:
        print(
            "Failed to retrieve authentication status from GitHub CLI:", file=sys.stderr
        )
        print(output, file=sys.stderr)
        raise RuntimeError("Failed to retrieve GitHub token")

    match = TOKEN_REGEX.search(output)
    if not match:
        print(
            (
                "Failed to find token in GitHub CLI output with regex"
                f" {TOKEN_REGEX.pattern!r}:"
            ),
            file=sys.stderr,
        )
        print(output, file=sys.stderr)
        raise RuntimeError("Failed to retrieve GitHub token")

    return match.groups()[0]


def _graphql(client: httpx.Client, query: str, variables: dict[str, object]):
    """
    Perform a GitHub GraphQL request.
    """
    github_token = get_github_token()

    response = (
        client.post(
            "https://api.github.com/graphql",
            json={"query": query, "variables": variables},
            headers={
                "Accept": "application/vnd.github+json",
                "Authorization": f"Bearer {github_token}",
            },
        )
        .raise_for_status()
        .json()
    )
    if response.get("errors"):
        raise RuntimeError(f"GraphQL server responded with error: {response['errors']}")
    return response


def parse_remote_url(remote_url: str) -> tuple[str, str]:
    """
    Parse a Git remote URL into owner and repository components.
    """
    ssh_prefix = "git@github.com:"
    if remote_url.startswith(ssh_prefix):
        owner_slash_repo = remote_url[len(ssh_prefix) :]
        owner, repo = owner_slash_repo.split("/")
        return owner, repo
    else:
        parts = remote_url.split("/")
        owner = parts[-2]
        repo = parts[-1]
        if repo.endswith(".git"):
            repo = repo[:-4]
        return owner, repo


def get_pull_requests_for_commits(
    owner, repo_name, commits: list[git.Commit]
) -> list[PullRequest]:
    """
    Retrieve the corresponding pull requests for a list of commits.

    Pull requests are retrieved in bulk, but GitHub enforces a page size of ~100 items
    so multiple HTTP requests may be made to retrieve all pull requests.

    This method [caches](`rooster._cache`) responses from GitHub to disk to avoid
    excessive requests on repeated invocations of `rooster`.
    """
    if not commits:
        return []

    pull_requests = []
    seen_commits = 0
    expected_commits = {str(commit.id) for commit in commits}

    query = textwrap.dedent(
        """
        query associatedPullRequest(
            $repo: String!, $owner: String!, $commit: String!, $after: String
        ) {
            repository(name: $repo, owner: $owner) {
                commit: object(expression: $commit) {
                    ... on Commit {
                        id
                        history(after: $after) {
                            nodes {
                                oid
                                associatedPullRequests(first: 1) {
                                    edges {
                                        node {
                                            title
                                            number
                                            author {
                                                login
                                            }
                                            labels(first: 10) {
                                                edges {
                                                    node {
                                                        name
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            pageInfo {
                                hasNextPage
                                endCursor
                            }
                        }
                    }
                }
            }
        }
        """
    )

    with cached_graphql_client() as client:
        first_commit = commits[0]
        page_start = None

        next_page = True
        while next_page and seen_commits < len(commits):
            response = _graphql(
                client,
                query,
                variables={
                    "owner": owner,
                    "repo": repo_name,
                    "commit": str(first_commit.id),
                    "after": page_start,
                },
            )

            response_commits = response["data"]["repository"]["commit"]["history"][
                "nodes"
            ]
            seen_commits += len(response_commits)
            for commit in response_commits:
                if commit["oid"] not in expected_commits:
                    continue

                associated_pull_requests = commit["associatedPullRequests"]["edges"]
                if not associated_pull_requests:
                    # TODO: Commits without pull requests should probably be included in
                    #       the changelog with a link to the commit and the commit
                    #       message with an opt-in
                    continue

                for item in associated_pull_requests:
                    pull_request = item["node"]
                    if not pull_request:
                        continue
                    labels = {
                        edge["node"]["name"] for edge in pull_request["labels"]["edges"]
                    }

                    pull_requests.append(
                        PullRequest(
                            title=pull_request["title"],
                            number=pull_request["number"],
                            labels=frozenset(labels),
                            author=pull_request["author"]["login"],
                            repo_name=repo_name,
                            repo_owner=owner,
                        )
                    )

            page_info = response["data"]["repository"]["commit"]["history"]["pageInfo"]
            next_page = page_info["hasNextPage"]
            page_start = page_info["endCursor"]

    return pull_requests


def get_release(repo_org: str, repo_name: str, tag_name: str) -> Release | None:
    github_token = get_github_token()
    try:
        release = (
            httpx.get(
                f"https://api.github.com/repos/{repo_org}/{repo_name}/releases/tags/{tag_name}",
                headers={
                    "Accept": "application/vnd.github+json",
                    "Authorization": f"Bearer {github_token}",
                },
            )
            .raise_for_status()
            .json()
        )
    except httpx.HTTPStatusError as exc:
        if exc.response.status_code == 404:
            return None
        raise

    return Release(
        id=release["id"],
        name=release["name"],
        tag=release["tag_name"],
        body=release["body"],
        draft=release["draft"],
        prerelease=release["prerelease"],
    )


def update_release_notes(
    repo_org: str, repo_name: str, release_id: str, content: str
) -> None:
    github_token = get_github_token()
    request = {"body": content}
    response = httpx.patch(
        f"https://api.github.com/repos/{repo_org}/{repo_name}/releases/{release_id}",
        headers={
            "Accept": "application/vnd.github+json",
            "Authorization": f"Bearer {github_token}",
        },
        json=request,
    )
    response.raise_for_status()
    return



from pathlib import Path

import tomllib
import typer

from rooster._changelog import (
    Changelog,
    VersionSection,
    generate_contributors,
)
from rooster._config import Config
from rooster._git import get_commits_between, get_remote_url
from rooster._github import get_pull_requests_for_commits, parse_remote_url
from rooster._pyproject import PyProjectError, update_pyproject_version
from rooster._versions import (
    BumpType,
    Version,
    bump_version,
    get_latest_version,
    get_previous_version,
    update_file_version,
    versions_from_git_tags,
)

app = typer.Typer()


@app.command()
def release(repo: Path = typer.Argument(default=Path(".")), bump: BumpType = None):
    """
    Create a new release.

    - Bumps the version in version files
    - Updates the changelog with changes from the latest version

    If no bump type is provided, the bump type will be detected based on the pull request labels.
    """
    config = Config.from_directory(repo)

    # Get the last release version
    versions = versions_from_git_tags(config, repo)
    last_version = get_latest_version(versions)
    if last_version:
        typer.echo(f"Found last version tag {last_version}.")
    else:
        typer.echo("It looks like there are no version tags for this project.")

    # Get the commits since the last release
    changes = list(get_commits_between(config, repo, last_version))
    since = "since last release" if last_version else "in the project"
    typer.echo(f"Found {len(changes)} commits {since}.")

    # Determine the GitHub repository to read
    owner, repo_name = parse_remote_url(get_remote_url(repo))

    # Collect pull requests corresponding to each commit
    typer.echo(f"Retrieving pull requests for changes from {owner}/{repo_name}...")
    pull_requests = get_pull_requests_for_commits(owner, repo_name, changes)

    # Collect the unique set of labels changed
    labels = set()
    for pull_request in pull_requests:
        labels.update(pull_request.labels)

    # Determine the version bump type based on the labels or user provided choice
    if bump:
        bump_type = bump
    else:
        bump_type = BumpType.patch
        for label in config.major_labels:
            if label in labels:
                typer.echo(f"Detected major version change due label {label}")
                bump_type = BumpType.major
                break

        for label in config.minor_labels:
            if label in labels:
                typer.echo(f"Detected minor version change due label {label}")
                bump_type = BumpType.minor
                break

        if bump_type == BumpType.patch:
            typer.echo(
                "Detected patch version change — did not see any major or minor labels"
            )

    new_version = bump_version(
        # If there is no previous version, start at 0.0.0
        last_version or Version("0.0.0"),
        bump_type,
    )
    typer.echo(f"Using new version {new_version}")

    # Generate a changelog entry for the version
    changelog_file = repo.joinpath(config.changelog_file)
    if not changelog_file.exists():
        changelog = Changelog.new()
        typer.echo("Creating new changelog file")
    else:
        # Load the existing changelog
        changelog = Changelog.from_file(changelog_file)

    section = VersionSection.from_pull_requests(
        document=changelog,
        config=config,
        version=new_version,
        pull_requests=pull_requests,
    )
    changelog.insert_version_section(section)
    changelog_file.write_text(changelog.to_markdown())
    typer.echo("Updated changelog")

    try:
        update_pyproject_version(repo.joinpath("pyproject.toml"), new_version)
        typer.echo("Updated version in pyproject.toml")
    except PyProjectError as exc:
        typer.echo(f"Failed to update pyproject.toml: {exc}")
        raise typer.Exit(1)

    for path in config.version_files:
        update_file_version(path, last_version, new_version)
        typer.echo(f"Updated version in {path}")


@app.command()
def changelog(
    repo: Path = typer.Argument(default=Path(".")),
    version: str = None,
    skip_existing: bool = False,
):
    """
    Generate the changelog for a version.

    If not provided, the version from the `pyproject.toml` file will be used.
    """
    config = Config.from_directory(repo)

    if version is None:
        # Get the version from the pyproject file
        pyproject_path = repo.joinpath("pyproject.toml")
        if not pyproject_path.exists():
            typer.echo(
                "No pyproject.toml file found; provide a version to generate an entry for."
            )
            raise typer.Exit(1)

        pyproject = tomllib.loads(pyproject_path.read_text())
        version = pyproject.get("project", {}).get("version")
        if not version:
            typer.echo(
                "No version found in the pyproject.toml; provide a version to generate an entry for."
            )
            raise typer.Exit(1)

        typer.echo(f"Found version {version}")

    # Parse the version
    version = Version(version)

    versions = versions_from_git_tags(config, repo)
    previous_version = get_previous_version(versions, version)
    if previous_version:
        typer.echo(f"Found previous version {previous_version}")

    changes = list(get_commits_between(config, repo, previous_version, version))
    if previous_version:
        typer.echo(f"Found {len(changes)} commits since {previous_version}")
    else:
        typer.echo(f"Found {len(changes)} commits")

    remote = get_remote_url(repo)
    if not remote:
        typer.echo(
            "No remote found; cannot retrieve pull requests to generate changelog entry"
        )
        raise typer.Exit(1)

    owner, repo_name = parse_remote_url(remote)

    typer.echo(f"Retrieving pull requests for changes from {owner}/{repo_name}")
    pull_requests = get_pull_requests_for_commits(owner, repo_name, changes)

    # Load the existing changelog
    changelog = Changelog.from_file(repo.joinpath(config.changelog_file))

    if skip_existing:
        existing_section = changelog.get_version_section(version)
        if existing_section:
            existing_entries = existing_section.all_entries()
            previous_count = len(pull_requests)
            pull_requests = [
                pr
                for pr in pull_requests
                if all(f"[#{pr.number}]" not in entry for entry in existing_entries)
            ]
            skipped = previous_count - len(pull_requests)
            if skipped:
                typer.echo(
                    f"Excluding {skipped} pull requests already in changelog entry for {version}"
                )

    section = VersionSection.from_pull_requests(
        document=changelog,
        config=config,
        version=version,
        pull_requests=pull_requests,
    )

    print(section.as_document().to_markdown())


@app.command()
def contributors(
    repo: Path = typer.Argument(default=Path(".")),
    version: str = None,
):
    """
    Generate a contributor list for a version.

    If not provided, the version from the `pyproject.toml` file will be used.

    Only includes contributors that authored a commit between the given version and one before it.
    """
    config = Config.from_directory(repo)
    if version is None:
        # Get the version from the pyproject file
        pyproject_path = repo.joinpath("pyproject.toml")
        if not pyproject_path.exists():
            typer.echo(
                "No pyproject.toml file found; provide a version to generate an entry for."
            )
            raise typer.Exit(1)

        pyproject = tomllib.loads(pyproject_path.read_text())
        version = pyproject["project"]["version"]
        typer.echo(f"Found version {version}")

    # Parse the version
    version = Version(version)

    versions = versions_from_git_tags(config, repo)
    previous_version = get_previous_version(versions, version)
    if previous_version:
        typer.echo(f"Found previous version {previous_version}")

    changes = list(get_commits_between(config, repo, previous_version))
    if previous_version:
        typer.echo(f"Found {len(changes)} commits since {previous_version}")
    else:
        typer.echo(f"Found {len(changes)} commits")

    remote = get_remote_url(repo)
    if not remote:
        typer.echo(
            "No remote found; cannot retrieve pull requests to generate changelog entry"
        )
        raise typer.Exit(1)

    owner, repo_name = parse_remote_url(remote)

    typer.echo(f"Retrieving pull requests for changes from {owner}/{repo_name}")
    pull_requests = get_pull_requests_for_commits(owner, repo_name, changes)

    config = Config.from_directory(repo)

    print(generate_contributors(pull_requests, config))


@app.command()
def backfill(
    repo: Path = typer.Argument(default=Path(".")),
    include_first: bool = False,
    clear: bool = False,
    start_version: str = None,
):
    """
    Regenerate the entire changelog.
    """
    config = Config.from_directory(repo)
    start_version = Version(start_version) if start_version else None

    # Generate a changelog entry for the version
    changelog_file = repo.joinpath(config.changelog_file)

    if clear or not changelog_file.exists():
        changelog = Changelog.new()
        typer.echo("Creating new changelog file")
    else:
        # Load the existing changelog
        changelog = Changelog.from_file(changelog_file)

    remote = get_remote_url(repo)
    if not remote:
        typer.echo(
            "No remote found; cannot retrieve pull requests to generate changelog entry"
        )
        raise typer.Exit(1)

    owner, repo_name = parse_remote_url(remote)
    typer.echo(f"Found remote repository {owner}/{repo_name}")

    versions = versions_from_git_tags(config, repo)
    for i, version in enumerate(versions):
        previous_version = versions[i - 1] if i > 0 else None

        if start_version:
            if version < start_version:
                continue

        changes = list(get_commits_between(config, repo, previous_version, version))
        if previous_version:
            typer.echo(
                f"Found {len(changes)} commits between {previous_version}...{version}"
            )
        else:
            if not include_first:
                continue

            typer.echo(f"Found {len(changes)} commits before {version}")

        typer.echo(f"Retrieving pull requests for {version}...")
        pull_requests = get_pull_requests_for_commits(owner, repo_name, changes)

        section = VersionSection.from_pull_requests(
            document=changelog,
            config=config,
            version=version,
            pull_requests=pull_requests,
        )
        changelog.insert_version_section(section)

    changelog_file.write_text(changelog.to_markdown())
    typer.echo("Updated changelog")



"""
Utilities for working with version numbers.
"""

from enum import Enum
from pathlib import Path
from typing import Any

import tomllib
from packaging.version import InvalidVersion, Version

from rooster._config import Config
from rooster._git import get_tags


class BumpType(Enum):
    major = "major"
    minor = "minor"
    patch = "patch"


def versions_from_git_tags(config: Config, repo: Path) -> list[Version]:
    """
    Get versions of the project from git tags.
    """
    tags = get_tags(config, repo)
    versions = parse_versions(tags)
    return versions


def parse_versions(version_strings: list[str]) -> list[Version]:
    """
    Parse version strings typed, validated objects.

    Invalid versions will be silently ignored.
    """
    versions = []
    for version in version_strings:
        try:
            versions.append(Version(version))
        except InvalidVersion:
            # Ignore tags that are not valid versions
            pass
    return versions


def get_latest_version(versions: list[Version]) -> Version | None:
    """
    Get the newest version from a collection of versions.
    """
    if not versions:
        return None
    return sorted(versions, reverse=True)[0]


def get_previous_version(versions: list[Version], version: Version) -> Version | None:
    """
    Get the version preceding a given version from a collection of versions.
    """
    if version not in versions:
        # Ensure the given version is included (but do not modify the given list)
        versions = [version] + versions
    versions = sorted(versions)
    index = versions.index(version) - 1
    if index < 0:
        return None
    return versions[index]


def bump_version(version: Version, bump_type: BumpType) -> Version:
    """
    Create a new version from a preceding one, increasing the given component.
    """
    # Pull the release section from the version and increment the appropriate number
    release = list(version.release)

    match bump_type:
        case BumpType.patch:
            release[2] += 1
        case BumpType.minor:
            release[1] += 1
            release[2] = 0
        case BumpType.major:
            release[0] += 1
            # Reset minor and patch versions
            release[1] = 0
            release[2] = 0
        case _:
            raise ValueError(f"Invalid bump type: {bump_type}")

    # Build a new  version string (`Version` is immutable)
    parts = []
    if version.epoch != 0:
        parts.append(f"{version.epoch}!")

    # Update release segment
    parts.append(".".join(str(x) for x in release))

    # We do not include other sections like dev/local/post since we are publishing
    # a new version. We could allow doing so in a separate function but then we
    # would need to construct the object again.
    return Version("".join(parts))


def update_file_version(path: Path, old_version: Version, new_version: Version) -> None:
    if path.name.lower() == "cargo.toml":
        update_toml_version(path, "package.version", old_version, new_version)
    elif path.name.lower() == "pyproject.toml":
        # Try the standard format
        try:
            return update_toml_version(
                path, "project.version", old_version, new_version
            )
        except KeyError:
            pass

        # Then try the Poetry format
        try:
            return update_toml_version(
                path, "tool.poetry.version", old_version, new_version
            )
        except KeyError:
            raise ValueError(
                f"Could not find `project.version` or `tool.poetry.version` at {path}."
            )

    elif path.name.lower().endswith(".md") or path.name.lower().endswith(".txt"):
        update_text_version(path, old_version, new_version)
    else:
        raise ValueError(
            f"Unsupported version update file {path.name}; expected 'Cargo.toml', 'pyproject.toml', '*.txt', or '*.md' file"
        )


def update_text_version(path: Path, old_version: Version, new_version: Version) -> None:
    """
    Update the version in a basic text file.
    """
    contents = path.read_text()
    path.write_text(contents.replace(str(old_version), str(new_version)))


def update_toml_version(
    path: Path, key: str, old_version: Version, new_version: Version
) -> None:
    """
    Update the version in a toml file.
    """
    contents = path.read_text()
    parsed = tomllib.loads(contents)

    # First check for the key to avoid replacing the wrong thing
    try:
        found_old_version = _get_nested_key(parsed, key)
    except KeyError:
        raise KeyError(f"{key} not found in {path}")

    # Ensure the contents matches the expected old version
    if Version(found_old_version) != old_version:
        raise Version(
            f"Mismatched version in {path}::{key}; expected {old_version} found {found_old_version}"
        )

    # Update with a string replacement to avoid reformatting the whole file
    contents = contents.replace(
        f'version = "{old_version}"', f'version = "{new_version}"', 1
    )

    # Confirm we updated the correct key
    new_parsed = tomllib.loads(contents)
    found_new_version = _get_nested_key(new_parsed, key)
    if found_new_version != str(new_version):
        raise RuntimeError(
            f"Failed safety check when updating version at {path}::{key}; expected {new_version} found {found_new_version}"
        )

    # Write the update
    path.write_text(contents)


def _get_nested_key(source: dict[str, Any], key: str):
    current = source
    for name in key.split("."):
        current = current[name]

    return current



from pathlib import Path
from typing import Generator

import pygit2 as git
from packaging.version import Version

from rooster._config import Config

TAG_PREFIX = "refs/tags/"


def get_tags(config: Config, target: Path) -> list[str]:
    repo = git.repository.Repository(target.absolute())

    references = repo.listall_references()
    return [
        _parse_tag_reference(config, ref)
        for ref in references
        if ref.startswith(TAG_PREFIX + config.version_tag_prefix)
    ]


def _parse_tag_reference(config: Config, reference: str) -> str:
    return reference[len(TAG_PREFIX + config.version_tag_prefix) :]


def get_commits_between(
    config: Config,
    target: Path,
    first_version: Version | None = None,
    second_version: Version | None = None,
) -> Generator[git.Commit, None, None]:
    """
    Yield all commits between two tags
    """
    repo = git.repository.Repository(target.absolute())
    first_commit = (
        repo.lookup_reference(
            TAG_PREFIX + config.version_tag_prefix + str(first_version)
        )
        .peel()
        .id
        if first_version is not None
        else None
    )
    second_commit = (
        repo.lookup_reference(
            TAG_PREFIX + config.version_tag_prefix + str(second_version)
        )
        .peel()
        .id
        if second_version is not None
        # TODO: Lookup main branch
        else repo.revparse_single("main").id
    )
    # Walk backwards from the second commit until we find the first commit
    for commit in repo.walk(second_commit):
        if commit.id == first_commit:
            break
        yield commit


def get_remote_url(target: Path, remote_name: str = "origin") -> str | None:
    repo = git.repository.Repository(target.absolute())
    names = {remote.name for remote in repo.remotes}
    if remote_name not in names:
        return None
    return repo.remotes[remote_name].url



"""
Utilities for interacting with `pyproject.toml` files.
"""
from pathlib import Path

import tomllib
from packaging.version import Version


class PyProjectError(Exception):
    """
    Error when reading a `pyproject.toml` file.
    """

    pass


def update_pyproject_version(path: Path, version: Version) -> None:
    """
    Update the version in a `pyproject.toml` file.
    """
    contents = path.read_text()
    parsed = tomllib.loads(contents)

    # We could consider supporting Poetry projects, but let's stick to the standards for now.
    if "project" not in parsed:
        raise PyProjectError("Missing `project` section.")

    if "version" not in parsed["project"]:
        raise PyProjectError("Missing `project.version` field.")

    old_version = parsed["project"]["version"]
    contents = contents.replace(f'version = "{old_version}"', f'version = "{version}"')
    path.write_text(contents)


def get_pyproject_version(path: Path) -> Version:
    """
    Read the current version from a `pyproject.toml` file.
    """
    contents = path.read_text()
    parsed = tomllib.loads(contents)
    return Version(parsed["project"]["version"])
